@@ -1,7 +1,7 @@
 """Classes used to build a schema.
 
 :organization: Logilab
-:copyright: 2003-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
+:copyright: 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
 :contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
 :license: General Public License version 2 - http://www.gnu.org/licenses
 """
@@ -22,14 +22,15 @@ __all__ = ('ObjectRelation', 'SubjectRelation', 'BothWayRelation',
            'SubjectRelation', 'ObjectRelation', 'BothWayRelation',
            'RichString', ) + tuple(BASE_TYPES)
 
-ETYPE_PROPERTIES = ('description', 'permissions',
+ETYPE_PROPERTIES = ('description', '__permissions__',
                     'meta') # XXX meta is deprecated
 # don't put description inside, handled "manualy"
-RTYPE_PROPERTIES = ('symetric', 'inlined', 'fulltext_container', 'permissions',
+RTYPE_PROPERTIES = ('symmetric', 'inlined', 'fulltext_container',
                     'meta') # XXX meta is deprecated
 RDEF_PROPERTIES = ('cardinality', 'constraints', 'composite',
                    'order',  'default', 'uid', 'indexed', 'uid',
-                   'fulltextindexed', 'internationalizable')
+                   'fulltextindexed', 'internationalizable',
+                   '__permissions__',)
 
 REL_PROPERTIES = RTYPE_PROPERTIES+RDEF_PROPERTIES + ('description',)
 
@@ -76,21 +77,32 @@ def _copy_attributes(fromobj, toobj, attributes):
             continue
         ovalue = getattr(toobj, attr, MARKER)
         if not ovalue is MARKER and value != ovalue:
+            rname = getattr(toobj, 'name', None) or toobj.__name__
             raise BadSchemaDefinition(
-                'conflicting values %r/%r for property %s of %s'
-                % (ovalue, value, attr, toobj))
+                'conflicting values %r/%r for property %s of relation %r'
+                % (ovalue, value, attr, rname))
         setattr(toobj, attr, value)
 
 def register_base_types(schema):
     for etype in BASE_TYPES:
-        edef = EntityType(name=etype)
-        schema.add_entity_type(edef).set_default_groups()
+        edef = EntityType(
+            name=etype,
+            # unused actually
+            # XXX add a group in read perms to satisfy schema constraints in cw
+            __permissions__={'read': ('users',), 'add': (), 'delete': (),
+                             'update': ()})
+        schema.add_entity_type(edef)
 
+# XXX use a "frozendict"
+_default_relperms = {'read': ('managers', 'users', 'guests',),
+                     'delete': ('managers', 'users'),
+                     'add': ('managers', 'users',)}
 
 class Relation(object):
     """Abstract class which have to be defined before the metadefinition
     meta-class.
     """
+    __permissions__ = MARKER
 
 # first class schema definition objects #######################################
 
@@ -99,6 +111,7 @@ class Definition(object):
 
     meta = MARKER
     description = MARKER
+    __permissions__ = MARKER
 
     def __init__(self, name=None):
         self.name = (name or getattr(self, 'name', None)
@@ -123,14 +136,59 @@ class Definition(object):
         """
         raise NotImplementedError()
 
+    @iclassmethod
+    def get_permissions(cls):
+        if cls.__permissions__ is MARKER:
+            return _default_relperms
+        return cls.__permissions__
 
-class metadefinition(type):
+    @classmethod
+    def set_permissions(cls, perms):
+        cls.__permissions__ = perms
+
+    @classmethod
+    def set_action_permissions(cls, action, actionperms):
+        permissions = cls.get_permissions().copy()
+        permissions[action] = actionperms
+        cls.__permissions__ = permissions
+
+
+class XXX_backward_permissions_compat(type):
+    stacklevel = 2
+    def __new__(mcs, name, bases, classdict):
+        if 'permissions' in classdict:
+            classdict['__permissions__'] = classdict.pop('permissions')
+            warn('[yams 0.27.0] permissions is deprecated, use __permissions__ instead (class %s)' % name,
+                 DeprecationWarning, stacklevel=mcs.stacklevel)
+        if 'symetric' in classdict:
+            classdict['symmetric'] = classdict.pop('symetric')
+            warn('[yams 0.27.0] symetric has been respelled symmetric (class %s)' % name,
+                 DeprecationWarning, stacklevel=mcs.stacklevel)
+        return super(XXX_backward_permissions_compat, mcs).__new__(mcs, name, bases, classdict)
+
+    # XXX backward compatiblity
+    def get_permissions(cls):
+        warn('[yams 0.27.0] %s.permissions is deprecated, use .__permissions__ instead'
+             % cls.__name__, DeprecationWarning, stacklevel=2)
+        return cls.__permissions__
+
+    def set_permissions(cls, newperms):
+        warn('[yams 0.27.0] %s.permissions is deprecated, use .__permissions__ instead'
+             % cls.__name__, DeprecationWarning, stacklevel=2)
+        cls.__permissions__ = newperms
+
+    permissions = property(get_permissions, set_permissions)
+
+
+class metadefinition(XXX_backward_permissions_compat):
     """Metaclass that builds the __relations__ attribute of
     EntityType's subclasses.
     """
+    stacklevel = 3
     def __new__(mcs, name, bases, classdict):
-        classdict['__relations__'] = rels = []
-        relations = {}
+
+        rels = classdict.setdefault('__relations__', [])
+        relations = dict((rdef.name, rdef) for rdef in rels)
         for rname, rdef in classdict.items():
             if isinstance(rdef, Relation):
                 # relation's name **must** be removed from class namespace
@@ -148,13 +206,16 @@ class metadefinition(type):
             _add_relation(defclass.__relations__, rdef, rname)
         # take base classes'relations into account
         for base in bases:
-            rels.extend(getattr(base, '__relations__', []))
+            for rdef in getattr(base, '__relations__', ()):
+                if not rdef.name in relations or not relations[rdef.name].override:
+                    rels.append(rdef)
+                else:
+                    relations[rdef.name].creation_rank = rdef.creation_rank
         # sort relations by creation rank
         defclass.__relations__ = sorted(rels, key=lambda r: r.creation_rank)
         return defclass
 
 
-
 class EntityType(Definition):
     # FIXME reader magic forbids to define a docstring...
     #"""an entity has attributes and can be linked to other entities by
@@ -171,13 +232,18 @@ class EntityType(Definition):
     #"""
 
     __metaclass__ = metadefinition
+    # XXX use a "frozendict"
+    __permissions__ = {
+        'read': ('managers', 'users', 'guests',),
+        'update': ('managers', 'owners',),
+        'delete': ('managers', 'owners'),
+        'add': ('managers', 'users',)
+        }
 
     def __init__(self, name=None, **kwargs):
         super(EntityType, self).__init__(name)
         _check_kwargs(kwargs, ETYPE_PROPERTIES)
-        _copy_attributes(attrdict(kwargs), self, ETYPE_PROPERTIES)
-        # if not hasattr(self, 'relations'):
-        #self.relations = list(self.__relations__)
+        self.__dict__.update(kwargs)
         self.specialized_type = self.__class__.__dict__.get('__specializes__')
 
     def __str__(self):
@@ -193,7 +259,10 @@ class EntityType(Definition):
         them to the `defined` dictionnary.
         """
         name = getattr(cls, 'name', cls.__name__)
-        assert name not in defined, "type '%s' was already defined" % name
+        assert cls is not defined.get(name), 'duplicate registration: %s' % name
+        assert name not in defined, \
+            "type '%s' was already defined here %s, new definition here %s" % \
+            (name, defined[name].__module__, cls)
         cls._defined = defined # XXX may be used later (eg .add_relation())
         defined[name] = cls
         for relation in cls.__relations__:
@@ -201,13 +270,22 @@ class EntityType(Definition):
 
     @classmethod
     def _ensure_relation_type(cls, relation):
+        """Check the type the relation
+
+        return False if the class is not yet finalized
+        (XXX raise excep instead ?)"""
         rtype = RelationType(relation.name)
         _copy_attributes(relation, rtype, RTYPE_PROPERTIES)
-        defined = cls._defined
-        if relation.name in defined:
-            _copy_attributes(rtype, defined[relation.name], RTYPE_PROPERTIES)
+        #assert hasattr(cls, '_defined'), "Type definition for %s not yet expanded. you can't register new type through it" % cls
+        if hasattr(cls, '_defined'):
+            defined = cls._defined
+            if relation.name in defined:
+                _copy_attributes(rtype, defined[relation.name], RTYPE_PROPERTIES)
+            else:
+                defined[relation.name] = rtype
+            return True
         else:
-            defined[relation.name] = rtype
+            return False
 
     @classmethod
     def expand_relation_definitions(cls, defined, schema):
@@ -243,12 +321,14 @@ class EntityType(Definition):
     def add_relation(cls, rdef, name=None):
         if name:
             rdef.name = name
-        cls._ensure_relation_type(rdef)
-        _add_relation(cls.__relations__, rdef, name)
-        if isinstance(rdef, RichString) and not rdef in cls._defined:
-            format_attr_name = (name or rdef.name) + '_format'
-            rdef = cls.get_relations(format_attr_name).next()
-            cls._ensure_relation_type(rdef)
+        if cls._ensure_relation_type(rdef):
+            _add_relation(cls.__relations__, rdef, name)
+            if isinstance(rdef, RichString) and not rdef in cls._defined:
+                format_attr_name = (name or rdef.name) + '_format'
+                rdef = cls.get_relations(format_attr_name).next()
+                cls._ensure_relation_type(rdef)
+        else:
+           _add_relation(cls.__relations__, rdef, name=name)
 
     @classmethod
     def insert_relation_after(cls, afterrelname, name, rdef):
@@ -267,7 +347,6 @@ class EntityType(Definition):
     def remove_relation(cls, name):
         for rdef in cls.get_relations(name):
             cls.__relations__.remove(rdef)
-            cls.__relations__.remove(rdef)
 
     @classmethod
     def get_relations(cls, name):
@@ -280,7 +359,9 @@ class EntityType(Definition):
 
 
 class RelationType(Definition):
-    symetric = MARKER
+    __metaclass__ = XXX_backward_permissions_compat
+
+    symmetric = MARKER
     inlined = MARKER
     fulltext_container = MARKER
 
@@ -288,9 +369,9 @@ class RelationType(Definition):
         """kwargs must have values in RTYPE_PROPERTIES"""
         super(RelationType, self).__init__(name)
         if kwargs.pop('meta', None):
-            warn('meta is deprecated', DeprecationWarning, stacklevel=2)
-        _check_kwargs(kwargs, RTYPE_PROPERTIES + ('description',))
-        _copy_attributes(attrdict(kwargs), self, RTYPE_PROPERTIES + ('description',))
+            warn('[yams 0.25] meta is deprecated', DeprecationWarning, stacklevel=2)
+        _check_kwargs(kwargs, RTYPE_PROPERTIES + ('description', '__permissions__'))
+        self.__dict__.update(kwargs)
 
     def __str__(self):
         return 'relation type %r' % self.name
@@ -305,10 +386,13 @@ class RelationType(Definition):
         if cls.__doc__ and not cls.description:
             cls.description = ' '.join(cls.__doc__.split())
         if name in defined:
-            _copy_attributes(cls, defined[name],
+            if defined[name].__class__ is not RelationType:
+                raise BadSchemaDefinition('duplicated relation type for %s'
+                                          % name)
+            # relation type created from a relation definition, override it
+            _copy_attributes(defined[name], cls,
                              REL_PROPERTIES + ('subject', 'object'))
-        else:
-            defined[name] = cls
+        defined[name] = cls
 
     @classmethod
     def expand_relation_definitions(cls, defined, schema):
@@ -328,14 +412,14 @@ class RelationDefinition(Definition):
     # FIXME reader magic forbids to define a docstring...
     #"""a relation is defined by a name, the entity types that can be
     #subject or object the relation, the cardinality, the constraints
-    #and the symetric property.
+    #and the symmetric property.
     #"""
 
     subject = MARKER
     object = MARKER
     cardinality = MARKER
     constraints = MARKER
-    symetric = MARKER
+    symmetric = MARKER
     inlined = MARKER
 
     def __init__(self, subject=None, name=None, object=None, **kwargs):
@@ -350,9 +434,13 @@ class RelationDefinition(Definition):
             self.object = self.__class__.object
         super(RelationDefinition, self).__init__(name)
         if kwargs.pop('meta', None):
-            warn('meta is deprecated', DeprecationWarning)
+            warn('[yams 0.25] meta is deprecated', DeprecationWarning)
+        if 'symetric' in kwargs:
+            warn('[yams 0.27.0] symetric has been respelled symmetric',
+                 DeprecationWarning, stacklevel=2)
+            kwargs['symmetric'] = kwargs.pop('symetric')
         _check_kwargs(kwargs, RDEF_PROPERTIES + ('description',))
-        _copy_attributes(attrdict(kwargs), self, RDEF_PROPERTIES + ('description',))
+        _copy_attributes(attrdict(**kwargs), self, RDEF_PROPERTIES + ('description',))
         if self.constraints:
             self.constraints = list(self.constraints)
 
@@ -367,6 +455,11 @@ class RelationDefinition(Definition):
         """
         name = getattr(cls, 'name', cls.__name__)
         rtype = RelationType(name)
+        if hasattr(cls, 'symetric'):
+            cls.symmetric = cls.symetric
+            del cls.symetric
+            warn('[yams 0.27.0] symetric has been respelled symmetric (class %s)' % name,
+                 DeprecationWarning)
         _copy_attributes(cls, rtype, RTYPE_PROPERTIES)
         if name in defined:
             _copy_attributes(rtype, defined[name], RTYPE_PROPERTIES)
@@ -405,10 +498,14 @@ class RelationDefinition(Definition):
             self.constraints = ()
         rschema = schema.rschema(name)
         if self.subject == '**' or self.object == '**':
-            warn('** is deprecated, use * (%s)' % rtype, DeprecationWarning)
+            warn('[yams 0.25] ** is deprecated, use * (%s)' % rtype, DeprecationWarning)
+        if self.__permissions__ is MARKER:
+            permissions = rtype.get_permissions()
+        else:
+            permissions = self.__permissions__
         for subj in _actual_types(schema, self.subject):
             for obj in _actual_types(schema, self.object):
-                rdef = RelationDefinition(subj, name, obj)
+                rdef = RelationDefinition(subj, name, obj, __permissions__=permissions)
                 _copy_attributes(self, rdef, RDEF_PROPERTIES + ('description',))
                 schema.add_relation_def(rdef)
 
@@ -426,7 +523,7 @@ def _actual_types(schema, etype):
 
 def _pow_etypes(schema):
     for eschema in schema.entities():
-        if eschema.is_final():
+        if eschema.final:
             continue
         yield eschema.type
 
@@ -450,15 +547,20 @@ class ObjectRelation(Relation):
         self.etype = etype
         if self.constraints:
             self.constraints = list(self.constraints)
+        self.override = kwargs.pop('override', False)
+        if 'symetric' in kwargs:
+            kwargs['symmetric'] = kwargs.pop('symetric')
+            warn('[yams 0.27.0] symetric has been respelled symmetric',
+                 DeprecationWarning, stacklevel=2)
         try:
             _check_kwargs(kwargs, REL_PROPERTIES)
         except BadSchemaDefinition, bad:
             # XXX (auc) bad field name + required attribute can lead there instead of schema.py ~ 920
-             bsd_ex = BadSchemaDefinition(('%s in relation to entity %r (also is %r defined ? (check two '
-                                           'lines above in the backtrace))') % (bad.args, etype, etype))
-             setattr(bsd_ex,'tb_offset',2)
-             raise bsd_ex
-        _copy_attributes(attrdict(kwargs), self, REL_PROPERTIES)
+            bsd_ex = BadSchemaDefinition(('%s in relation to entity %r (also is %r defined ? (check two '
+                                          'lines above in the backtrace))') % (bad.args, etype, etype))
+            bsd_ex.tb_offset = 2
+            raise bsd_ex
+        self.__dict__.update(kwargs)
 
     def __repr__(self):
         return '%(name)s %(etype)s' % self.__dict__
@@ -552,7 +654,7 @@ class RichString(String):
 # various derivated classes with some predefined values XXX deprecated
 
 class MetaEntityType(EntityType):
-    permissions = {
+    __permissions__ = {
         'read':   ('managers', 'users', 'guests',),
         'add':    ('managers',),
         'update': ('managers', 'owners',),
@@ -563,7 +665,7 @@ class MetaUserEntityType(EntityType):
     pass
 
 class MetaRelationType(RelationType):
-    permissions = {
+    __permissions__ = {
         'read':   ('managers', 'users', 'guests',),
         'add':    ('managers',),
         'delete': ('managers',),
@@ -574,7 +676,7 @@ class MetaUserRelationType(RelationType):
 
 class MetaAttributeRelationType(RelationType):
     # just set permissions to None so default permissions are set
-    permissions = MARKER
+    __permissions__ = MARKER
 
 
 __all__ += ('MetaEntityType', 'MetaUserEntityType',
