@@ -4,7 +4,7 @@ Use either a sql derivated language for entities and relation definitions
 files or a direct python definition file.
 
 :organization: Logilab
-:copyright: 2004-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
+:copyright: 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
 :contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
 :license: General Public License version 2 - http://www.gnu.org/licenses
 """
@@ -13,7 +13,7 @@ __docformat__ = "restructuredtext en"
 import sys
 import types
 from os import listdir
-from os.path import exists, join, splitext, basename
+from os.path import exists, join, splitext, basename, abspath
 from warnings import warn
 
 from logilab.common.testlib import mock_object
@@ -81,6 +81,7 @@ class PyFileReader(object):
             modname, module = self._loaded[filepath]
         except KeyError:
             modname, module = self.exec_file(filepath)
+        processed_object = set()
         for name, obj in vars(module).items():
             if name.startswith('_'):
                 continue
@@ -88,8 +89,11 @@ class PyFileReader(object):
                 isdef = issubclass(obj, buildobjs.Definition)
             except TypeError:
                 continue
-            if isdef and obj.__module__ == modname:
-                self.loader.add_definition(self, obj)
+            if (not isdef) or obj.__module__ != modname or \
+                obj in processed_object:
+                continue
+            self.loader.add_definition(self, obj)
+            processed_object.add(obj)
 
     def import_erschema(self, ertype, schemamod=None, instantiate=True):
         warn('import_erschema is deprecated, use explicit import once schema '
@@ -123,7 +127,7 @@ class PyFileReader(object):
         if modname in sys.modules:
             module = sys.modules[modname]
             # NOTE: don't test raw equality to avoid .pyc / .py comparisons
-            assert module.__file__.startswith(filepath), (filepath, module.__file__)
+            assert abspath(module.__file__).startswith(abspath(filepath)), (filepath, module.__file__)
         else:
             # XXX until bw compat is gone, put context into builtins to allow proper
             # control of deprecation warning
@@ -172,6 +176,8 @@ class PyFileReader(object):
             sys.modules[modname] = module
             if package:
                 setattr(sys.modules[package], modname.split('.')[-1], module)
+        if hasattr(module, 'post_build_callback'):
+            self.loader.post_build_callbacks.append(module.post_build_callback)
         self._loaded[filepath] = (modname, module)
         return self._loaded[filepath]
 
@@ -200,11 +206,12 @@ class SchemaLoader(object):
         self.defined = {}
         self.loaded_files = []
         self._pyreader = PyFileReader(self)
+        self.post_build_callbacks = []
         sys.modules[__name__].context = self
         # ensure we don't have an iterator
         directories = tuple(directories)
         try:
-            files = self._load_definition_files(directories)
+            self._load_definition_files(directories)
             try:
                 schema = self._build_schema(name, register_base_types,
                                             construction_mode=construction_mode,
@@ -248,14 +255,17 @@ class SchemaLoader(object):
             if isinstance(definition, type):
                 definition = definition()
             definition.expand_relation_definitions(self.defined, schema)
+        # call 'post_build_callback' functions found in schema modules
+        for cb in self.post_build_callbacks:
+            cb(schema)
+        # optionaly remove relation types without definitions
         if remove_unused_rtypes:
-            # remove relation types without definitions
             for rschema in schema.relations():
-                if not rschema.rdefs():
+                if not rschema.rdefs:
                     schema.del_relation_type(rschema)
         # set permissions on entities and relations
         for erschema in schema.entities() + schema.relations():
-            erschema.set_default_groups()
+            erschema.check_permission_definitions()
         schema.infer_specialization_rules()
         return schema
 
